from __future__ import annotations

from dataclasses import KW_ONLY, dataclass
from typing import Literal, final

from uniserde import JsonDoc

import rio.docs

from .fundamental_component import FundamentalComponent

__all__ = [
    "Popup",
    "PopupOpenOrCloseEvent",
]


@final
@rio.docs.mark_constructor_as_private
@dataclass
class PopupOpenOrCloseEvent:
    """
    The argument for the `on_open_or_close` handler of a `Popup`.
    """

    is_open: bool


@final
class Popup(FundamentalComponent):
    """
    A container which floats above other components.

    Popups are containers which float above the page when open. This allows you
    to keep your app clean by default, but present additional information or
    controls when needed.

    They take two children: The `anchor` is always visible and positions the
    popup. The `content` is located inside the popup and is only visible when
    the popup is open.

    The location popups appear at can be customized using the `direction`,
    `alignment` and `gap` attributes. Popups wil do their best to honor those
    settings, but deviate if necessary to ensure they don't go off-screen.


    ## Attributes

    `anchor`: A component which is always visible and positions the popup.

    `content`: A component which is only visible when the popup is open.

    `direction`: The direction into which the popup opens.

    `alignment`: The alignment of the popup within the anchor. If the popup
        opens to the left or right, this is the vertical alignment, with `0`
        being the top and `1` being the bottom. If the popup opens to the top or
        bottom, this is the horizontal alignment, with `0` being the left and
        `1` being the right. Has no effect if the popup opens centered.

    `gap`: How much space to leave between the popup and the anchor. Has no
        effect popup opens centered.

    `is_open`: Whether the popup is currently open.

    `on_open_or_close`: Triggered when the popup is opened or closed.
    """

    # TODO: Add example to docstring

    anchor: rio.Component
    content: rio.Component
    _: KW_ONLY
    color: rio.ColorSet = "neutral"
    direction: Literal["left", "top", "right", "bottom", "center"] = "center"
    alignment: float = 0.5
    gap: float = 0.8
    is_open: bool = False
    on_open_or_close: rio.EventHandler[PopupOpenOrCloseEvent] = None

    def _validate_delta_state_from_frontend(self, delta_state: JsonDoc) -> None:
        if not set(delta_state) <= {"is_open"}:
            raise AssertionError(
                f"Frontend tried to change `{type(self).__name__}` state: {delta_state}"
            )

    async def _call_event_handlers_for_delta_state(
        self, delta_state: JsonDoc
    ) -> None:
        # Trigger on_open_or_close event
        try:
            is_open = delta_state["is_open"]
        except KeyError:
            pass
        else:
            assert isinstance(is_open, bool), is_open
            await self.call_event_handler(
                self.on_open_or_close,
                PopupOpenOrCloseEvent(is_open),
            )

    def _custom_serialize(self) -> JsonDoc:
        return {
            "color": self.session.theme._serialize_colorset(self.color),
        }


Popup._unique_id = "Popup-builtin"



import asyncio
import contextlib
import functools
import inspect
import json
import types
from collections.abc import (
    AsyncGenerator,
    Callable,
    Container,
    Iterable,
    Iterator,
    Mapping,
)
from typing import Any, TypeVar

import ordered_set
from uniserde import Jsonable, JsonDoc

import rio.global_state
from rio.app_server import AppServer
from rio.components.root_components import (
    FundamentalRootComponent,
    HighLevelRootComponent,
)

T = TypeVar("T")
C = TypeVar("C", bound=rio.Component)


async def _make_awaitable(value: T = None) -> T:
    return value


class MockApp:
    def __init__(
        self,
        session: rio.Session,
        user_settings: JsonDoc = {},
    ) -> None:
        self.session = session
        self.outgoing_messages: list[JsonDoc] = []

        self._first_refresh_completed = asyncio.Event()

        self._responses = asyncio.Queue[JsonDoc]()
        self._responses.put_nowait(
            {
                "websiteUrl": "https://unit.test",
                "preferredLanguages": [],
                "userSettings": user_settings,
                "windowWidth": 1920,
                "windowHeight": 1080,
                "timezone": "America/New_York",
                "decimalSeparator": ".",
                "thousandsSeparator": ",",
                "prefersLightTheme": True,
            }
        )

    async def _send_message(self, message_text: str) -> None:
        message = json.loads(message_text)

        self.outgoing_messages.append(message)

        if "id" in message:
            self._responses.put_nowait(
                {
                    "jsonrpc": "2.0",
                    "id": message["id"],
                    "result": None,
                }
            )

        if message["method"] == "updateComponentStates":
            self._first_refresh_completed.set()

    async def _receive_message(self) -> Jsonable:
        return await self._responses.get()

    @property
    def dirty_components(self) -> Container[rio.Component]:
        return set(self.session._dirty_components)

    @property
    def last_updated_components(self) -> set[rio.Component]:
        return set(self.last_component_state_changes)

    @property
    def last_component_state_changes(
        self,
    ) -> Mapping[rio.Component, Mapping[str, object]]:
        for message in reversed(self.outgoing_messages):
            if message["method"] == "updateComponentStates":
                delta_states: dict = message["params"]["deltaStates"]  # type: ignore
                return {
                    self.session._weak_components_by_id[
                        int(component_id)
                    ]: delta
                    for component_id, delta in delta_states.items()
                    if int(component_id) != self.session._root_component._id
                }

        return {}

    def get_root_component(self) -> rio.Component:
        sess = self.session

        high_level_root = sess._root_component
        assert isinstance(
            high_level_root, HighLevelRootComponent
        ), high_level_root

        low_level_root = sess._weak_component_data_by_component[
            high_level_root
        ].build_result
        assert isinstance(
            low_level_root, FundamentalRootComponent
        ), low_level_root

        scroll_container = low_level_root.content
        assert isinstance(
            scroll_container, rio.ScrollContainer
        ), scroll_container

        return scroll_container.content

    def get_components(self, component_type: type[C]) -> Iterator[C]:
        root_component = self.get_root_component()

        for component in root_component._iter_component_tree():
            if type(component) is component_type:
                yield component  # type: ignore

    def get_component(self, component_type: type[C]) -> C:
        try:
            return next(self.get_components(component_type))
        except StopIteration:
            raise AssertionError(f"No component of type {component_type} found")

    def get_build_output(
        self,
        component: rio.Component,
        type_: type[C] | None = None,
    ) -> C:
        result = self.session._weak_component_data_by_component[
            component
        ].build_result

        if type_ is not None:
            assert (
                type(result) is type_
            ), f"Expected {type_}, got {type(result)}"

        return result  # type: ignore

    async def refresh(self) -> None:
        await self.session._refresh()


@contextlib.asynccontextmanager
async def create_mockapp(
    build: Callable[[], rio.Component] = lambda: rio.Text("hi"),
    *,
    app_name: str = "mock-app",
    running_in_window: bool = False,
    user_settings: JsonDoc = {},
    default_attachments: Iterable[object] = (),
    use_ordered_dirty_set: bool = False,
) -> AsyncGenerator[MockApp, None]:
    app = rio.App(
        build=build,
        name=app_name,
        default_attachments=tuple(default_attachments),
    )
    app_server = AppServer(
        app,
        debug_mode=False,
        running_in_window=running_in_window,
        validator_factory=None,
        internal_on_app_start=None,
    )

    # Emulate the process of creating a session as closely as possible
    fake_request: Any = types.SimpleNamespace(
        url="https://unit.test",
        base_url="https://unit.test",
        headers={"accept": "text/html"},
    )
    await app_server._serve_index(fake_request, "")

    [[session_token, session]] = app_server._active_session_tokens.items()

    if use_ordered_dirty_set:
        session._dirty_components = ordered_set.OrderedSet(
            session._dirty_components
        )  # type: ignore

    mock_app = MockApp(session, user_settings=user_settings)

    fake_websocket: Any = types.SimpleNamespace(
        client="1.2.3.4",
        accept=lambda: _make_awaitable(),
        send_text=mock_app._send_message,
        receive_json=mock_app._receive_message,
    )

    test_task = asyncio.current_task()
    assert test_task is not None

    async def serve_websocket():
        try:
            await app_server._serve_websocket(fake_websocket, session_token)
        except asyncio.CancelledError:
            pass
        except Exception as error:
            test_task.cancel(
                f"Exception in AppServer._serve_websocket: {error}"
            )
        else:
            test_task.cancel(
                "AppServer._serve_websocket exited unexpectedly. An exception"
                " must have occurred in the `init_coro`."
            )

    server_task = asyncio.create_task(serve_websocket())

    await mock_app._first_refresh_completed.wait()
    try:
        yield mock_app
    finally:
        server_task.cancel()


def enable_component_instantiation(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        def build():
            result = func(*args, **kwargs)
            assert not inspect.isawaitable(
                result
            ), "The test function must not be async"

            return rio.Text("")

        app = rio.App(build=build)
        app_server = AppServer(
            app_=app,
            debug_mode=False,
            running_in_window=False,
            validator_factory=None,
            internal_on_app_start=None,
        )
        session = rio.Session(
            app_server,
            "<a fake session token>",
        )
        session._decimal_separator = "."
        session._thousands_separator = ","
        session._send_message = _fake_send_message
        session._receive_message = _fake_receive_message

        rio.global_state.currently_building_session = session
        session._root_component = HighLevelRootComponent(
            build, lambda: rio.Text("")
        )

        rio.global_state.currently_building_component = session._root_component
        try:
            session._root_component.build()
        finally:
            rio.global_state.currently_building_component = None
            rio.global_state.currently_building_session = None

    return wrapper


async def _fake_send_message(message: Jsonable) -> None:
    pass


async def _fake_receive_message() -> Jsonable:
    while True:
        await asyncio.sleep(100000)



from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import *  # type: ignore

import openai


@dataclass
class ChatMessage:
    """
    A simple storage class containing all the information needed for a single
    chat message.
    """

    role: Literal["user", "assistant"]
    timestamp: datetime
    text: str


@dataclass
class Conversation:
    """
    The start of the show. This class contains a list of messages and can
    connect to OpenAI to generate smart responses to the user's messages.
    """

    # The entire message history
    messages: List[ChatMessage] = field(default_factory=list)

    async def respond(self, client: openai.AsyncOpenAI) -> ChatMessage:
        """
          Creates an AI generated response for this conversation and appends it
          to the messages list. Also returns the new message.

        ## Raises

          `ValueError` if the most recent message is not by the user.
        """

        # Make sure the last message was by the user
        if not self.messages or self.messages[-1].role != "user":
            raise ValueError("The most recent message must be by the user")

        # Convert all messages to the format needed by the API
        api_messages: list[Any] = [
            {
                "role": "system",
                "content": "You are a helpful assistant. Format your response in markdown, for example by using **bold**, and _italic_ amongst others.",
            }
        ] + [
            {
                "role": message.role,
                "content": message.text,
            }
            for message in self.messages
        ]

        # Generate a response
        api_response = await client.chat.completions.create(
            model="gpt-3.5-turbo",
            messages=api_messages,
            max_tokens=500,
        )

        assert isinstance(api_response.choices[0].message.content, str)

        response = ChatMessage(
            role="assistant",
            timestamp=datetime.now(tz=timezone.utc),
            text=api_response.choices[0].message.content,
        )

        # Append the message and return it as well
        self.messages.append(response)

        return response



from __future__ import annotations

import pathlib
from typing import Literal, final

from uniserde import JsonDoc

import rio

from .. import assets, color
from ..common import EventHandler
from .fundamental_component import KeyboardFocusableFundamentalComponent

__all__ = ["MediaPlayer"]


@final
class MediaPlayer(KeyboardFocusableFundamentalComponent):
    """
    Plays audio and video.

    `MediaPlayer` plays back audio and video files. It can play local files and
    URLs.

    Note that the `MediaPlayer` component doesn't reserve any space for itself,
    it simply makes do with the space it is given by its parent component.


    ## Attributes

    `media`: The media to play. This can be a file path, URL, or bytes.

    `media_type`: The mime type of the media file. May help the browser play the
        file correctly.

    `loop`: Whether to automatically restart from the beginning when the
        playback ends.

    `autoplay`: Whether to start playing the media automatically, without
        requiring the user to press "Play".

    `controls`: Whether to display controls like a Play/Pause button, volume
        slider, etc.

    `muted`: Whether to mute the audio.

    `volume`: The volume to play the audio at. 1.0 is the native volume;
        larger numbers increase the volume, smaller numbers decrease it.

    `background`: The background fill. This is visible when playing audio or
        when the video doesn't use up all the available space.

    `on_playback_end`: An event handler to call when the media finishes
        playing.

    `on_error`: An event handler to call when an error occurs, for example if
        the file format isn't supported.


    ## Examples

    A minimal example of a `MediaPlayer` playing a media file from the internet:

    ```python
    rio.MediaPlayer(rio.URL("https://example.com/example_video.mp4"))
    ```

    You can use a local file as well:

    ```python
    from pathlib import Path

    rio.MediaPlayer(Path("example_video.mp4"))
    ```
    """

    media: pathlib.Path | rio.URL | bytes
    media_type: str | None
    loop: bool
    autoplay: bool
    controls: bool
    muted: bool
    volume: float
    background: rio.Fill
    on_playback_end: EventHandler[[]]
    on_error: EventHandler[[]]

    def __init__(
        self,
        media: pathlib.Path | rio.URL | bytes,
        *,
        media_type: str | None = None,
        loop: bool = False,
        autoplay: bool = False,
        controls: bool = True,
        muted: bool = False,
        volume: float = 1.0,
        background: rio.FillLike = color.Color.TRANSPARENT,
        on_playback_end: EventHandler[[]] = None,
        on_error: EventHandler[[]] = None,
        key: str | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        width: float | Literal["natural", "grow"] = "natural",
        height: float | Literal["natural", "grow"] = "natural",
        align_x: float | None = None,
        align_y: float | None = None,
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.media = media
        self.media_type = media_type
        self.loop = loop
        self.autoplay = autoplay
        self.controls = controls
        self.muted = muted
        self.volume = volume
        self.background = rio.Fill._try_from(background)
        self.on_playback_end = on_playback_end
        self.on_error = on_error

    def _custom_serialize(self) -> JsonDoc:
        media_asset = assets.Asset.new(self.media, self.media_type)
        return {
            "mediaUrl": media_asset._serialize(self.session),
            "reportError": self.on_error is not None,
            "reportPlaybackEnd": self.on_playback_end is not None,
        }

    def _validate_delta_state_from_frontend(self, delta_state: JsonDoc) -> None:
        if not set(delta_state) <= {"muted", "volume"}:
            raise AssertionError(
                f"Frontend tried to change `{type(self).__name__}` state: {delta_state}"
            )

    async def _on_message(self, message: JsonDoc) -> None:  # type: ignore
        if message["type"] == "playbackEnd":
            await self.call_event_handler(self.on_playback_end)
        elif message["type"] == "error":
            await self.call_event_handler(self.on_error)
        else:
            await super()._on_message(message)


MediaPlayer._unique_id = "MediaPlayer-builtin"



from pathlib import Path

import numpy as np
import sentence_transformers.util
from sentence_transformers import SentenceTransformer


def scan_for_icons(path: Path) -> list[str]:
    result = []

    for fpath in path.glob("**/*.svg"):
        name = fpath.stem.replace("-", " ")
        result.append(name)

    return result


model = SentenceTransformer(
    "all-mpnet-base-v2",
    # "multi-qa-MiniLM-L6-cos-v1",
)


query = "alarm"
query_embedding = model.encode("How big is London")

hay = scan_for_icons(
    Path("/home/jakob/.cache/rio/extracted-icon-sets/material")
)
hay_embeddings = model.encode(hay)

# Find the 5 most similar entries
similarities = sentence_transformers.util.dot_score(
    query_embedding, hay_embeddings
)
similarities = similarities.ravel()

topk = np.argsort(similarities)
topk = topk[-5:]


for ii in topk:
    print(f"{hay[ii]}: {similarities[ii]}")



import threading
import time
from typing import *  # type: ignore

from . import run_models

try:
    import webview  # type: ignore
except ImportError:
    if TYPE_CHECKING:
        import webview  # type: ignore
    else:
        webview = None


class WebViewWorker:
    def __init__(
        self,
        *,
        push_event: Callable[[run_models.Event], None],
        debug_mode: bool,
        url: str,
    ) -> None:
        self.push_event = push_event
        self.debug_mode = debug_mode
        self.url = url

        # If running, this is the webview window
        self.window: Optional[webview.Window] = None

    def run(self) -> None:
        """
        Starts the worker and blocks until the window is closed or
        `request_stop` is called. This function must be called from the main
        thread.
        """
        assert webview is not None
        assert self.window is None, "Already running"

        # Make sure this was called from the main thread.
        assert (
            threading.current_thread() is threading.main_thread()
        ), "Must be called from the main thread"

        # Create the window
        self.window = webview.create_window(
            # TODO: Get the app's name, if possible
            "Rio (debug)" if self.debug_mode else "Rio",
            url=self.url,
        )
        webview.start()

        # If we've reached this point, the window must have been closed.
        self.window = None

        # If the webview window is closed, tell the Arbiter about it.
        self.push_event(run_models.StopRequested())

    def request_stop(self) -> None:
        """
        Makes the worker stop soon after called. May be called multiple times,
        or even before startup has finished.
        """

        while self.window is not None:
            # Try to close the window
            try:
                self.window.destroy()

            # Closing can fail if the window hasn't entirely opened yet. In that
            # case try again in a bit.
            except KeyError:  # TODO: Why key error?
                time.sleep(0.05)  # TODO

            # Success. Make sure to forget the window.
            else:
                self.window = None



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass
from typing import Any, Literal, final

from uniserde import Jsonable

import rio.docs

from .fundamental_component import KeyboardFocusableFundamentalComponent

__all__ = [
    "KeyEventListener",
    "HardwareKey",
    "SoftwareKey",
    "KeyDownEvent",
    "KeyUpEvent",
    "KeyPressEvent",
]


HardwareKey = Literal[
    "unknown",
    # Function keys
    "f1",
    "f2",
    "f3",
    "f4",
    "f5",
    "f6",
    "f7",
    "f8",
    "f9",
    "f10",
    "f11",
    "f12",
    "f13",
    "f14",
    "f15",
    "f16",
    "f17",
    "f18",
    "f19",
    "f20",
    "f21",
    "f22",
    "f23",
    "f24",
    # Digits
    "0",
    "1",
    "2",
    "3",
    "4",
    "5",
    "6",
    "7",
    "8",
    "9",
    # Letters
    "a",
    "b",
    "c",
    "d",
    "e",
    "f",
    "g",
    "h",
    "i",
    "j",
    "k",
    "l",
    "m",
    "n",
    "o",
    "p",
    "q",
    "r",
    "s",
    "t",
    "u",
    "v",
    "w",
    "x",
    "y",
    "z",
    # Punctuation
    "comma",
    "period",
    "semicolon",
    "single-quote",
    "backquote",
    "space",
    # Brackets
    "bracket-left",
    "bracket-right",
    # Math
    "plus",
    "minus",
    "asterisk",
    "slash",
    "equal",
    # Numpad
    "numpad-0",
    "numpad-1",
    "numpad-2",
    "numpad-3",
    "numpad-4",
    "numpad-5",
    "numpad-6",
    "numpad-7",
    "numpad-8",
    "numpad-9",
    "numpad-decimal",
    "numpad-plus",
    "numpad-minus",
    "numpad-asterisk",
    "numpad-slash",
    "numpad-enter",
    "numpad-equal",
    "numpad-comma",
    # Arrow keys
    "arrow-up",
    "arrow-down",
    "arrow-left",
    "arrow-right",
    # Modifiers
    "left-alt",
    "right-alt",
    "left-control",
    "right-control",
    "left-meta",
    "right-meta",
    "left-meta",
    "right-meta",
    "left-shift",
    "right-shift",
    # Toggles
    "caps-lock",
    "num-lock",
    "scroll-lock",
    "fn-lock",
    # Browser
    "browser-back",
    "browser-favorites",
    "browser-forward",
    "browser-home",
    "browser-refresh",
    "browser-search",
    "browser-stop",
    # Media
    "media-play-pause",
    "media-track-next",
    "media-track-previous",
    "media-stop",
    "media-select",
    "media-eject",
    "media-eject",  # Alias of MediaEject
    "volume-down",
    "volume-up",
    "volume-down",  # Alias of VolumeDown
    "volume-up",  # Alias of VolumeUp
    "volume-mute",
    # Apps
    "launch-app-1",
    "launch-app-2",
    "launch-mail",
    # Power
    "power",
    "sleep",
    "wake-up",
    # Clipboard
    "copy",
    "cut",
    "paste",
    # Languages
    "lang-1",
    "lang-2",
    "lang-3",
    "lang-4",
    # Misc
    "escape",
    "tab",
    "enter",
    "delete",
    "insert",
    "home",
    "end",
    "page-up",
    "page-down",
    "pause",
    "print-screen",
    "context-menu",
    "help",
    "backspace",
    "convert",
    "non-convert",
    "backslash",
    "left-backslash",
    "undo",
    "kana-mode",
    "intl-ro",
    "intl-yen",
    "fn",
    "again",
    "props",
    "select",
    "execute",
    "find",
    "cancel",
    "redo",
    "zoom-in",
    "zoom-out",
    "clear",
    "brightness-up",
    "brightness-down",
]

SoftwareKey = Literal[
    "unknown",
    # Modifiers
    "alt",
    "control",
    "meta",
    "meta",
    "shift",
    "alt-graph",
    "caps-lock",
    "num-lock",
    "scroll-lock",
    "fn",
    "fn-lock",
    "super",
    "hyper",
    "symbol",
    "symbol-lock",
    # Whitespace
    "enter",
    "tab",
    "space",
    # Navigation
    "arrow-down",
    "arrow-left",
    "arrow-right",
    "arrow-up",
    "end",
    "home",
    "page-down",
    "page-up",
    # Editing
    "backspace",
    "clear",
    "copy",
    "cursor-select",
    "cut",
    "delete",
    "erase-eof",
    "extend-selection",
    "insert",
    "paste",
    "redo",
    "undo",
    # UI
    "accept",
    "again",
    "attention",
    "cancel",
    "context-menu",
    "escape",
    "execute",
    "find",
    "finish",
    "help",
    "pause",
    "play",
    "props",
    "select",
    "zoom-in",
    "zoom-out",
    # Device
    "brightness-down",
    "brightness-up",
    "eject",
    "log-off",
    "power",
    "power-off",
    "print-screen",
    "hibernate",
    "standby",
    "wake-up",
    # IME and composition
    "all-candidates",
    "alphanumeric",
    "code-input",
    "compose",
    "convert",
    "dead",
    "final-mode",
    "group-first",
    "group-last",
    "group-next",
    "group-previous",
    "mode-change",
    "next-candidate",
    "non-convert",
    "previous-candidate",
    "process",
    "single-candidate",
    # Korean
    "hangul-mode",
    "hanja-mode",
    "junja-mode",
    # Japanese
    "eisu",
    "hankaku",
    "hiragana",
    "hiragana-katakana",
    "kana-mode",
    "kanji-mode",
    "katakana",
    "romaji",
    "zenkaku",
    "zenkaku-hanaku",
    # Function
    "f1",
    "f2",
    "f3",
    "f4",
    "f5",
    "f6",
    "f7",
    "f8",
    "f9",
    "f10",
    "f11",
    "f12",
    "f13",
    "f14",
    "f15",
    "f16",
    "f17",
    "f18",
    "f19",
    "f20",
    "f21",  # TODO: According to MDN they only go up to F20
    "f22",
    "f23",
    "f24",
    "soft-1",
    "soft-2",
    "soft-3",
    "soft-4",
    # Phone
    "app-switch",
    "call",
    "camera",
    "camera-focus",
    "end-call",
    "go-back",
    "go-home",
    "headset-hook",
    "last-number-redial",
    "notification",
    "manner-mode",
    "voice-dial",
    # Multimedia
    "channel-down",
    "channel-up",
    "media-fast-forward",
    "media-pause",
    "media-play",
    "media-play-pause",
    "media-record",
    "media-rewind",
    "media-stop",
    "media-track-next",
    "media-track-previous",
    # Audio control
    # TODO: I stopped checking MDN at this point, this is purely Copilot now
    "audio-balance-left",
    "audio-balance-right",
    "audio-bass-down",
    "audio-bass-boost-down",
    "audio-bass-boost-toggle",
    "audio-bass-boost-up",
    "audio-bass-up",
    "audio-fader-front",
    "audio-fader-rear",
    "audio-surround-mode-next",
    "audio-treble-down",
    "audio-treble-up",
    "audio-volume-down",
    "audio-volume-mute",
    "audio-volume-up",
    "microphone-toggle",
    "microphone-volume-down",
    "microphone-volume-mute",
    "microphone-volume-up",
    # TV control
    "tv",
    "tv-3d-mode",
    "tv-antenna-cable",
    "tv-audio-description",
    "tv-audio-description-mix-down",
    "tv-audio-description-mix-up",
    "tv-contents-menu",
    "tv-data-service",
    "tv-input",
    "tv-input-component-1",
    "tv-input-component-2",
    "tv-input-composite-1",
    "tv-input-composite-2",
    "tv-input-hdmi-1",
    "tv-input-hdmi-2",
    "tv-input-hdmi-3",
    "tv-input-hdmi-4",
    "tv-input-vga-1",
    "tv-media-context",
    "tv-network",
    "tv-number-entry",
    "tv-power",
    "tv-radio-service",
    "tv-satellite",
    "tv-satellite-bs",
    "tv-satellite-cs",
    "tv-satellite-toggle",
    "tv-terrestrial-analog",
    "tv-terrestrial-digital",
    "tv-timer",
    # Media controller
    "avr-input",
    "avr-power",
    "color-f0-red",
    "color-f1-green",
    "color-f2-yellow",
    "color-f3-blue",
    "color-f4-grey",
    "color-f5-brown",
    "closed-caption-toggle",
    "dimmer",
    "display-swap",
    "dvr",
    "exit",
    "favorite-clear-0",
    "favorite-clear-1",
    "favorite-clear-2",
    "favorite-clear-3",
    "favorite-recall-0",
    "favorite-recall-1",
    "favorite-recall-2",
    "favorite-recall-3",
    "favorite-store-0",
    "favorite-store-1",
    "favorite-store-2",
    "favorite-store-3",
    "guide",
    "guide-next-day",
    "guide-previous-day",
    "info",
    "instant-replay",
    "link",
    "list-program",
    "live-content",
    "lock",
    "media-apps",
    "media-audio-track",
    "media-last",
    "media-skip-backward",
    "media-skip-forward",
    "media-step-backward",
    "media-step-forward",
    "media-top-menu",
    "navigate-in",
    "navigate-next",
    "navigate-out",
    "navigate-previous",
    "next-favorite-channel",
    "next-user-profile",
    "on-demand",
    "pairing",
    "pin-p-down",
    "pin-p-move",
    "pin-p-toggle",
    "pin-p-up",
    "play-speed-down",
    "play-speed-reset",
    "play-speed-up",
    "random-toggle",
    "rc-low-battery",
    "record-speed-next",
    "rf-bypass",
    "scan-channels-toggle",
    "screen-mode-next",
    "settings",
    "split-screen-toggle",
    "stb-input",
    "stb-power",
    "subtitle",
    "teletext",
    "video-mode-next",
    "wink",
    "zoom-toggle",
    # Speech recognition
    "speech-correction-list",
    "speech-input-toggle",
    # Document
    "close",
    "new",
    "open",
    "print",
    "save",
    "spell-check",
    "mail-forward",
    "mail-reply",
    "mail-send",
    # Application launcher
    "launch-calculator",
    "launch-calendar",
    "launch-contacts",
    "launch-mail",
    "launch-media-player",
    "launch-music-player",
    "launch-my-computer",
    "launch-phone",
    "launch-screen-saver",
    "launch-spreadsheet",
    "launch-web-browser",
    "launch-web-cam",
    "launch-word-processor",
    "launch-application-1",
    "launch-application-2",
    "launch-application-3",
    "launch-application-4",
    "launch-application-5",
    "launch-application-6",
    "launch-application-7",
    "launch-application-8",
    "launch-application-9",
    "launch-application-10",
    "launch-application-11",
    "launch-application-12",
    "launch-application-13",
    "launch-application-14",
    "launch-application-15",
    "launch-application-16",
    # Browser
    "browser-back",
    "browser-favorites",
    "browser-forward",
    "browser-home",
    "browser-refresh",
    "browser-search",
    "browser-stop",
    # Numeric keypad
    "decimal",
    "key-11",
    "key-12",
    "asterisk",
    "plus",
    "slash",
    "minus",
    "separator",
]

ModifierKey = Literal["alt", "control", "meta", "shift"]

_MODIFIERS = ("control", "shift", "alt", "meta")


@dataclass(frozen=True)
class _KeyUpDownEvent:
    """
    ## Attributes

    `hardware_key`: The name of the physical button on the keyboard.

    `software_key`: The name of the button. Depending on the user's keyboard
        layout, this may differ from the `hardware_key`. (For example, if the
        physical keyboard has a QWERTY layout but the OS is configured to use
        Dvorak.)

    `text`: The text that pressing this button produces. For example, Shift+1
        produces the text "!". If the button doesn't produce any text, like F1
        for example, this will be an empty string.

    `modifiers`: The names of the modifier keys (control, shift, alt, meta) that
        were pressed when the event occurred.
    """

    hardware_key: HardwareKey
    software_key: SoftwareKey | str
    text: str
    modifiers: frozenset[ModifierKey]

    @classmethod
    def _from_json(cls, json_data: dict[str, Any]):
        return cls(
            hardware_key=json_data["hardwareKey"],
            software_key=json_data["softwareKey"],
            text=json_data["text"],
            modifiers=frozenset(json_data["modifiers"]),
        )

    def __str__(self) -> str:
        keys = [key for key in _MODIFIERS if key in self.modifiers]
        keys.append(self.software_key)

        return " + ".join(keys)


@final
@rio.docs.mark_constructor_as_private
class KeyDownEvent(_KeyUpDownEvent):
    """
    The argument for the `on_key_down` handler of a `KeyEventListener`.
    """


@final
@rio.docs.mark_constructor_as_private
class KeyUpEvent(_KeyUpDownEvent):
    """
    The argument for the `on_key_up` handler of a `KeyEventListener`.
    """


@final
@rio.docs.mark_constructor_as_private
class KeyPressEvent(_KeyUpDownEvent):
    """
    The argument for the `on_key_press` handler of a `KeyEventListener`.
    """


@final
class KeyEventListener(KeyboardFocusableFundamentalComponent):
    """
    Calls an event handler when a key is pressed or released.

    `KeyEventListener` is a container for a single child component. It listens
    for keyboard events and calls corresponding event handlers when they occur.

    Keep in mind that `KeyEventListener` will only report events that have not
    been handled by a child component. For example, typing into a `TextInput`
    will not trigger a `KeyEventListener`.


    ## Attributes

    `content`: The child component.

    `on_key_down`: A function to call when a key is pressed down.

    `on_key_up`: A function to call when a key is released.

    `on_key_press`: A function to call repeatedly while a key is held down.
    """

    content: rio.Component
    _: KW_ONLY
    on_key_down: rio.EventHandler[KeyDownEvent] = None
    on_key_up: rio.EventHandler[KeyUpEvent] = None
    on_key_press: rio.EventHandler[KeyPressEvent] = None

    def _custom_serialize(self) -> dict[str, Jsonable]:
        return {
            "reportKeyDown": self.on_key_down is not None,
            "reportKeyUp": self.on_key_up is not None,
            "reportKeyPress": self.on_key_press is not None,
        }

    async def _on_message(self, msg: Any) -> None:
        # Parse the message
        assert isinstance(msg, dict), msg

        msg_type: str = msg["type"]
        assert isinstance(msg_type, str), msg_type

        # Dispatch the correct event
        if msg_type == "KeyDown":
            await self.call_event_handler(
                self.on_key_down,
                KeyDownEvent._from_json(msg),
            )

        elif msg_type == "KeyUp":
            await self.call_event_handler(
                self.on_key_up,
                KeyUpEvent._from_json(msg),
            )

        elif msg_type == "KeyPress":
            await self.call_event_handler(
                self.on_key_press,
                KeyPressEvent._from_json(msg),
            )

        else:
            raise ValueError(
                f"{__class__.__name__} encountered unknown message: {msg}"
            )

        # Refresh the session
        await self.session._refresh()


KeyEventListener._unique_id = "KeyEventListener-builtin"
